/**
 * printPS.cpp - Bitcoin poscript generator
 *
 * Copyright (C)2015 The bitgen developers <bitgen@openmailbox.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "printPS.h"

#include <iostream>
#include <string>
#include <list>

#include "util.h"
#include "qrenc.h"
#include "currencySpecific.h"
#include "url.h"
#include "ExtendedKey.h"
#include "StaticXor.h"

#include "psWriter.h"








std::pair<std::string, std::string> splitInTwoParts(const std::string& str)
{
	const int middlePos = str.size() / 2;
	const std::string start = str.substr(0, middlePos);
	const std::string end = str.substr(middlePos);
	
	const std::pair<std::string, std::string> ret(start, end);
	return ret;
}


std::list<std::string> splitInNParts(const std::string& str, const int maxSize = 65)
{
	//std::cout << "Will split in parts" << std::endl;
	
	std::list<std::string> theParts;
	
	const int partMaxSize(maxSize);
	const int parts = (str.size() / partMaxSize) + 1;

	//std::cout << "Parts: " << parts << std::endl;

	
	int currentPos(0);
	for(int i = 0 ; i < parts ; i++)
	{
		//std::cout << "Iterating: " << i << std::endl;
		const int nextPos = currentPos + partMaxSize;
		const std::string partStr = str.substr(currentPos, partMaxSize);
		if(partStr.empty())
		{
			break;
		}
		
		theParts.push_back(partStr);
		currentPos += partMaxSize;
	}
	
	
	
	//const int middlePos = str.size() / 2;
	return theParts;
}




std::string PostscriptGenerator::publicRoot(const HierarcialInfo& hierarcialInfo,
	const PublicPostscriptInfo& pubInfo, const std::string& label) const
{
	PSTextWriter textWriter(false, margin);

	textWriter.addPs("Big_font setfont\n");

	textWriter.writeLn((lang == LANG_EN) ? "PUBLIC ROOT" : "TEILEN");	
	textWriter.addPs("Text_font setfont\n");

	if(!label.empty())
	{
		textWriter.writeLn("Label:", label);
	}

	textWriter.writeLn((lang == LANG_EN) ? "Date:" : "Datum:", pubInfo.thetime);
	textWriter.writeLn("Signature:", hierarcialInfo.signature);
	textWriter.writeLn("Name:", pubInfo.accountName);

	std::string myText;
	myText += "Use this public root key to generate ";
	const std::string currencyName = getCurrencyName();
	myText += currencyName;
	myText += " addresses.";		

	textWriter.writeLn(myText);

	std::string myCommand;
	myCommand += "Command: bitgen";
	
	const std::string currencyShortName = getCurrencyShortName();
	
	if(currencyShortName != "btc")
	{
		myCommand += "-";
		myCommand += currencyShortName;
	}
	
	myCommand += " genPub root_";
	myCommand += intToString(info.getNewNum());
	myCommand += "_";
	myCommand += currencyShortName;
	myCommand += "_";
	myCommand += hierarcialInfo.signature;
	myCommand += "_pub.pubroot";

	textWriter.writeLn(myCommand);

	textWriter.addNewline(3);
	textWriter.writeLn("Public point X");

	const std::string pubXStart = hierarcialInfo.pubX.substr(0, 32);
	const std::string pubXEnd = hierarcialInfo.pubX.substr(32);

	textWriter.writeLn(pubXStart);
	textWriter.writeLn(pubXEnd);

	textWriter.writeLn("Public point Y");

	const std::pair<std::string, std::string> pubYParts = splitInTwoParts(hierarcialInfo.pubY);
	const std::string pubYStart = pubYParts.first;		//hierarcialInfo.pubY.substr(0, 32);
	const std::string pubYEnd = pubYParts.second; 		//hierarcialInfo.pubY.substr(32);

	textWriter.writeLn(pubYStart);
	textWriter.writeLn(pubYEnd);


	textWriter.addPs("Text_font setfont\n");

	const std::string ret = textWriter.getPs();

	//ret += lineBetween(margin, 418, 540, 418);
	
	return ret;	
}







std::string PostscriptGenerator::privateRoot(const std::string& privIndex, const std::string& signature) const
{
	PSTextWriter textWriter(false, margin);
	
	textWriter.addPs("Big_font setfont\n");
	textWriter.writeLn("PRIVATE ROOT");

	textWriter.addPs("Text_font setfont\n");

	textWriter.writeLn("Use this private root key to find the private root for a particular bitcoin address.");
	textWriter.writeLn("You must first generate the public part explained above.");
	textWriter.writeLn("Command:");

	std::string command;
	
	const std::string currencyShortName = getCurrencyShortName();
	
	command += "bitgen";
	if(currencyShortName != "btc")
	{
		command += "-";
		command += currencyShortName;
	}
	
	command += " genPriv root_";
	command += intToString(info.getNewNum());
	command += "_";
	command += currencyShortName;
	command += "_";
	command += signature;
	command += "_priv.privroot PRIVINDEXFILE.privindex";
	
	textWriter.writeLn(command);

	textWriter.writeLn("");

	textWriter.writeLn("Private index:");
		
	const std::string privStart = privIndex.substr(0, 32);
	const std::string privEnd = privIndex.substr(32);

	textWriter.writeLn(privStart);
	textWriter.writeLn(privEnd);


	std::string ret = textWriter.getPs();
		
	return ret;
}




std::string PostscriptGenerator::generatePaymentRequest(
						const PublicPostscriptInfo& pubInfo,
						const std::string& amount, const std::string& message,
						const std::string& thetime,
						const std::string paymentId) const
{
	PSTextWriter textWriter(false, margin, 140);

	textWriter.addPs("Big_font setfont\n");


	textWriter.writeLn("INVOICE");
	
	textWriter.addPs("Text_font setfont\n");

	textWriter.writeLn("Invoice date:", pubInfo.thetime);


	if(!amount.empty())
	{
		const std::string currencyShortName = getCurrencyShortName();
		
		textWriter.writeLn("Requested amount:", amount + " " + currencyShortName);		
	}

	if(!pubInfo.accountName.empty())
	{
		textWriter.writeLn("Label:", pubInfo.accountName);		
	}

	if(!message.empty())
	{
		textWriter.writeLn("Message:", message);		
	}

	if(!thetime.empty())
	{
		textWriter.writeLn("Last payment date:", thetime);		
	}

	const std::string currencyName = getCurrencyName();
	
	std::string payString;	
	payString += "Pay to ";
	payString += currencyName;
	payString += " address ";

	if(!paymentId.empty())
	{
		textWriter.writeLn("Payment id:", paymentId);		
	}

	textWriter.addNewline(4);
	textWriter.writeLn(payString);


	if(pubInfo.pubkey.size() < 50)
	{
		textWriter.writeLn(pubInfo.pubkey);		
	}
	else
	{		
		const std::pair<std::string, std::string> pubParts = splitInTwoParts(pubInfo.pubkey);		
		textWriter.writeLn(pubParts.first);
		textWriter.writeLn(pubParts.second);		
	}
	

	textWriter.addNewline(8);

	const std::string myEncStr = createEncString(pubInfo, amount, message);
	const QRInfo qrinfo = QRInfo::qrencode((const unsigned char*)myEncStr.c_str(), myEncStr.size());

	textWriter.addQR(qrinfo, "pub address", pubInfo.pubkey, 15.0);

	textWriter.addPs("Text_font setfont\n");

	return textWriter.getPs();
}







/*
std::string PostscriptGenerator::publicKeyPart(const PublicPostscriptInfo& pubInfo) const
{
	std::string ret;

	ret += "Titel_font setfont\n";
	
	ret += textAt(margin,       780, pubInfo.pubkey);

	ret += "Text_font setfont\n";

	std::string myText;
	myText += "Generated by ";
	myText += applicationName;
	myText += " ";
	myText += versionNumber;
	
	ret += textAt(margin,       760, myText);

	if(!pubInfo.thetime.empty())
	{
		std::string dateStr;
		dateStr += (lang == LANG_EN) ? "Generate date" : "Generieren Datum";
		dateStr += ": ";
		dateStr += pubInfo.thetime;
		
		ret += textAt(margin,       740, dateStr);		
	}

	if(!pubInfo.accountName.empty())
	{
		ret += textAt(margin,       720, "Keypair name: " + pubInfo.accountName);		
	}

	ret += textAt(margin,       680, (lang == LANG_EN) ? "Bitcoin pubkey " : "Bitcoin-Adresse");

	ret += textAt(margin,       660, pubInfo.pubkey);

	const std::string myEncStr = createEncString(pubInfo, "amount", "message");

	const QRInfo qrinfo = QRInfo::qrencode((const unsigned char*)myEncStr.c_str(), myEncStr.size());

	ret += qrCode(qrinfo, margin, 480, 15.0);

	ret += "Big_font setfont\n";


	ret += textAt(270,       610, "SHARE");
	
	ret += textAt(270,       560, "Encrypt &");
	ret += textAt(270,       510, "verify signature");


	ret += lineBetween(margin, 418, 540, 418);
	
	ret += "Text_font setfont\n";

	return ret;
}
*/





QRInfo PostscriptRawGenerator::qrEncodeAddress(const std::string& pubkeyVal, const std::string& accountName, 
	const bool useAlphaNumeric)
{
	const std::string urlPrefix = getUrlPrefix();
	std::string encStr(urlPrefix + pubkeyVal);

/*
	if(useAlphaNumeric && accountName.empty())
	{
		std::cout << "Will encode alpha numeric: " << encStr << std::endl;
		const std::string upperStr = toUpper(encStr);
		std::cout << "Will encode alpha numeric upper: " << upperStr << std::endl;
		
		const QRInfo qrinfo = QRInfo::qrencodeAlfaNumeric(upperStr.c_str());
		return qrinfo;		
	}
*/
	
	if(!accountName.empty())
	{
		encStr += "?label=";		
		encStr += urlEncode(accountName);
	}

	const QRInfo qrinfo = QRInfo::qrencode((const unsigned char*)encStr.c_str(), encStr.size());
	return qrinfo;
}



QRInfo PostscriptRawGenerator::qrEncodeExtendedAddress(const std::string& pub)
{
	const QRInfo qrinfo = QRInfo::qrencode((const unsigned char*)pub.c_str(), pub.size());
	return qrinfo;
}









std::string PostscriptGenerator::publicMnemonic(const PublicExtendedPostscriptInfo& pubInfo, const bool useSmallDescription, 
												const int numWords,
												const std::string& derivedXpub) const
{
	PSTextWriter textWriter(false, margin);
	
	const std::string currencyName = getCurrencyName();
	
	const QRInfo qrinfo = isExtended() ? PostscriptRawGenerator::qrEncodeExtendedAddress(pubInfo.pubkey) :
										PostscriptRawGenerator::qrEncodeAddress(pubInfo.pubkey, pubInfo.accountName);

		

	textWriter.addPs("Titel_font setfont\n");

	textWriter.writeLn("Mnemonic");
	
	textWriter.addPs("Text_font setfont\n");

	textWriter.write("Version:", applicationName + " " + versionNumber);

	if(currencySpecific || !derivedXpub.empty())
	{
		textWriter.writeSec2("Currency:", currencyName);
		textWriter.addNewline(1);		
	}


	if(!pubInfo.thetime.empty())
	{
		textWriter.writeLn("Date:", pubInfo.thetime);		
	}

	textWriter.write("Name:", pubInfo.accountName);			


	if(!pubInfo.nodeName.empty())
	{
		textWriter.writeSec2("Node name:", pubInfo.nodeName);			
	}

	textWriter.addNewline(1);

	textWriter.writeLn("Num words:", intToString(numWords));


	if(!pubInfo.pubkey.empty())
	{		
		const std::pair<std::string, std::string> pubYParts = splitInTwoParts(pubInfo.pubkey);
		
		textWriter.writeLn(pubYParts.first);
		textWriter.writeLn(pubYParts.second);
	}


	textWriter.addNewline(2);

	if(useSmallDescription)
	{
		textWriter.addPs("Titel_font setfont\n");
	}
	else
	{
		textWriter.addPs("Big_font setfont\n");		
	}
	
	//writeSec2
	textWriter.writeTab(6, (lang == LANG_EN) ? "SHARE" : "TEILEN");

	textWriter.addPs("Text_font setfont\n");	

	textWriter.addNewline(6);


	if(!pubInfo.pubkey.empty())
	{
		const QRInfo qrinfo = PostscriptRawGenerator::qrEncodeExtendedAddress(pubInfo.pubkey);

		textWriter.addQR(qrinfo, "xpub", "", 17.0);		
	}

	if(!derivedXpub.empty())
	{
		const std::string derivationStr = DerivationPath::getCurrencyDerivationPath();
		const std::string infoStr = "derived xpub " + derivationStr;
		
		const QRInfo qrinfo = PostscriptRawGenerator::qrEncodeExtendedAddress(derivedXpub);

		textWriter.addQR(qrinfo, infoStr, "", 17.0, true); //200);
		
	}

	textWriter.addPs("Text_font setfont\n");	

	std::string ret = textWriter.getPs();

	ret += PostscriptLineGenerator::lineBetween(margin, 418, 540, 418);

	return ret;
}






std::string PostscriptRawGenerator::qrCode(const QRInfo& qrinfo, const int x, const int y, const double scale) //const
{
	std::string ret;
	
	const double scalefactor = scale / double(qrinfo.height());

	
	ret += "gsave ";	
	ret += intToString(x);
	ret += " ";	
	ret += intToString(y);
	ret += " translate ";	
	ret += doubleToString(scalefactor);	
	ret += " ";	
	ret += doubleToString(scalefactor);			
	ret += " scale\n";	
	ret += generateQR(qrinfo);
	ret += "grestore\n";
	
	return ret;
}


std::pair<std::string, std::string> splitAtPos(const std::string& str, const int numChar)
{
	if(str.size() >= numChar)
	{
		const std::string pub1 = str.substr(0, numChar);
		const std::string pub2 = str.substr(numChar);
		return std::pair<std::string, std::string>(pub1, pub2);
	}
	else
	{
		return std::pair<std::string, std::string>(str, "");		
	}
}



/*
	if(isExtended())
	{
		ret += textAt(margin,       yPos, "Derivation path:");
		ret += textAt(margin + 100, yPos, "m");
		yPos -= 20;
		
		const std::pair<std::string, std::string> pubParts = splitInTwoParts(pubInfo.pubkey.toString());
		
		const std::string pubLeft = pubParts.first;
		const std::string pubRight = pubParts.second;

		ret += textAt(margin,       yPos, pubLeft);
		yPos -= 20;
		
		ret += textAt(margin,       yPos, pubRight);	
		yPos -= 20;
	}
	else
	{
	*/

/*
	if(!pubInfo.accountName.empty())
	{
		ret += textAt(margin,       yPos, "Name:");
		ret += textAt(margin + 70, yPos, pubInfo.accountName);		
		yPos -= 20;		
	}
*/



std::string PostscriptGenerator::publicPart(const PublicPostscriptInfo& pubInfo, const bool useSmallText, const bool useBigInfoFont) const
{	
	const std::string currencyName = getCurrencyName();

	PSTextWriter textWriter(false, margin, 80);

	textWriter.addPs("Titel_font setfont\n");
	
	const std::string pubAddressStr = pubInfo.pubkey;
	const std::string displayAdr;

	textWriter.writeLn((lang == LANG_EN) ? currencyName + " address " + displayAdr : currencyName + "-Adresse " + displayAdr);				

	textWriter.addPs("Text_font setfont\n");

	textWriter.write("Version:", applicationName + " " + versionNumber);

	if(!isExtended() || currencySpecific || isElectrum())
	{
		textWriter.writeSec2("Currency:", currencyName);
	}

	textWriter.newLine();

	if(!pubInfo.thetime.empty())
	{
		textWriter.writeLn((lang == LANG_EN) ? "Date:" : "Datum:", pubInfo.thetime);
	}


	textWriter.write("Name:", pubInfo.accountName);


	if(!pubInfo.computerName.empty())
	{
		textWriter.writeSec2("Node name:", pubInfo.computerName);			
	}
		
	textWriter.writeLn("");


	//if(!pubInfo.accountName.empty() || !pubInfo.computerName.empty())
	//{
	//	if(!pubInfo.accountName.empty())
	//	{
	//	}
	//}



	if(pubInfo.displayHashOfTime)
	{
		textWriter.writeLn("Time hash:", hexString(ripemd160(pubInfo.thetime)).substr(0, 8));
	}
	else
	{			
		std::pair<std::string, std::string> splitPub = splitAtPos(pubInfo.pubkey, 47);
		const std::string pub1 = splitPub.first;
		const std::string pub2 = splitPub.second;
						
		textWriter.writeLn((lang == LANG_EN) ? "Address:" : "Adresse:", pub1);

		if(!pub2.empty())
		{
			textWriter.writeLn("", pub2);			
		}			
	}



	textWriter.addNewline(3);

	if(useBigInfoFont)
	{
		textWriter.addPs("Big_font setfont\n");		
		
		textWriter.writeSec2((lang == LANG_EN) ? "SHARE" : "TEILEN", "");
		textWriter.addNewline(2);
		textWriter.writeSec2((lang == LANG_EN) ? "Load&verify" : "Laden & prüfen", "");
	}
	else
	{
		textWriter.writeSec2("", (lang == LANG_EN) ? "SHARE" : "TEILEN");
		textWriter.addNewline(1);
		textWriter.writeSec2("", (lang == LANG_EN) ? "Load & verify" : "Laden & prüfen");
		textWriter.addNewline(1);
	}


	textWriter.addPs("Text_font setfont\n");



	textWriter.addNewline(2);

	if(!pubInfo.displayHashOfTime)	//Address specified
	{
		const QRInfo qrinfo = isExtended() ? PostscriptRawGenerator::qrEncodeExtendedAddress(pubInfo.pubkey) :
										PostscriptRawGenerator::qrEncodeAddress(pubInfo.pubkey, pubInfo.accountName);
		const std::string pubLegacyStr = "pub";

		/*isExtended() ? 470 : 500,*/
		textWriter.addQR(qrinfo, (isExtended() ? "xpub" : pubLegacyStr), "", 15.0);
	}

	textWriter.addNewline(2);


	textWriter.addPs("Text_font setfont\n");

	const std::string addDateBlock = (getCurrencyShortName() == "zec") ? "Block:______________" : ((lang == LANG_EN) ? "Date:______________" : "Datum:__________");


	textWriter.write((lang == LANG_EN) ? "Amount added:____________" : "Menge zugesetzt:____________", 0);
	textWriter.write(addDateBlock, 200);
	textWriter.write((lang == LANG_EN) ? "Spent" : "Verbraucht", 350);

	textWriter.box(480, 20);

	if(useSmallText)
	{
		textWriter.addPs("Titel_font setfont\n");
	}
	else
	{
		textWriter.addPs("Big_font setfont\n");		
	}

	std::string ret = textWriter.getPs();
	
	ret += PostscriptLineGenerator::lineBetween(margin, 418, 540, 418);
		
	ret += "Text_font setfont\n";

	return ret;
}







	//if(getCurrencyShortName() == "zec")
	//{
	//	//ret += textAt(margin + 200, 460, "Block:______________");
	//}
	//else
	//{
	//	//ret += textAt(margin + 200, 460, (lang == LANG_EN) ? "Date:______________" : "Datum:__________");		
	//}


	//ret += textAt(270 + (useSmallText ? 200 : 0), publicLinenoNumToYpos(8) /*610*/, (lang == LANG_EN) ? "SHARE" : "TEILEN");

	//ret += textAt(270 + (useSmallText ? 200 : 0), publicLinenoNumToYpos(10) /*560*/, (lang == LANG_EN) ? "Load & verify" : "Laden & prüfen");


std::string PostscriptGenerator::publicPartExtended(const PublicExtendedPostscriptInfo& pubInfo, /*const bool useSmallText, */const bool useBigInfoFont) const
{
	const std::string currencyName = getCurrencyName();

	PSTextWriter textWriter(false, margin);

	textWriter.addPs("Titel_font setfont\n");

	if(isExtended())
	{
		std::string extAdr;
		if(addrType == ADDRTYPE_EXTENDED)
		{
			extAdr += "Extended address";
		}
		else
		{
			extAdr += "Extended electrum address";			
		}
		
		if(!pubInfo.accountName.empty())
		{
			extAdr += ": ";
			extAdr += pubInfo.accountName;
		}

		textWriter.writeLn(extAdr);		
	}
	else
	{
		const std::string displayAdr = pubInfo.displayHashOfTime ? "" : pubInfo.pubkey;
				
		textWriter.writeLn((lang == LANG_EN) ? currencyName + " address " + displayAdr : 
													currencyName + "-Adresse " + displayAdr);
	}
	

	textWriter.addPs("Text_font setfont\n");

	textWriter.write("Version:", applicationName + " " + versionNumber);	

	if(!isExtended() || currencySpecific || isElectrum())
	{
		textWriter.writeSec2("Currency:", currencyName);		
	}

	textWriter.writeLn("");

	if(!pubInfo.thetime.empty())
	{
		textWriter.writeLn((lang == LANG_EN) ? "Date:" : "Generieren Datum:", pubInfo.thetime);			
	}



	if(!pubInfo.accountName.empty() || !pubInfo.nodeName.empty())
	{
		if(!pubInfo.accountName.empty())
		{
			textWriter.write("Account:", pubInfo.accountName);				
		}

		if(!pubInfo.nodeName.empty())
		{
			textWriter.writeSec2("Node name:", pubInfo.nodeName);
		}
		
		textWriter.writeLn("");
	}



	if(isExtended())
	{		
		textWriter.writeLn("Deriv path:", "m");

		const std::pair<std::string, std::string> pubParts = splitInTwoParts(pubInfo.pubkey);
		
		const std::string pubLeft = pubParts.first;
		const std::string pubRight = pubParts.second;

		textWriter.writeLn(pubLeft);
		textWriter.writeLn(pubRight);		
	}
	else
	{
		if(pubInfo.displayHashOfTime)
		{
			textWriter.writeLn("Time hash:", hexString(ripemd160(pubInfo.thetime)).substr(0, 8));			
		}
		else
		{
			textWriter.writeLn((lang == LANG_EN) ? currencyName + " address " : currencyName + " adresse ", pubInfo.pubkey);			
		}

		textWriter.write((lang == LANG_EN) ? "Amount added:____________" : "Menge zugesetzt:____________", "");
		textWriter.write((lang == LANG_EN) ? "Date:__________" : "Datum:______", "");
		textWriter.write((lang == LANG_EN) ? "Spent" : "Verbraucht", "");

		textWriter.box(450, 20);
		textWriter.writeLn("");
	}

	textWriter.addNewline(2);


	if(useBigInfoFont)
	{
		textWriter.addPs("Big_font setfont\n");
	}
	else
	{
		textWriter.addPs("Titel_font setfont\n");
	}
	

	textWriter.writeSec2((lang == LANG_EN) ? "SHARE" : "TEILEN", "");
	textWriter.addNewline(2);	
	textWriter.writeSec2((lang == LANG_EN) ? "Load & verify" : "Laden & prüfen", "");
	textWriter.addNewline(2);	


	textWriter.addNewline(2);

	if(!pubInfo.displayHashOfTime)	//Address specified
	{
		const QRInfo qrinfo = isExtended() ? PostscriptRawGenerator::qrEncodeExtendedAddress(pubInfo.pubkey) :
										PostscriptRawGenerator::qrEncodeAddress(pubInfo.pubkey, pubInfo.accountName);

		textWriter.addQR(qrinfo, (isExtended() ? "xpub" : "pub"), "", 15.0);		
		
		
		if(!isExtended() && (getCurrencyShortName() == "bch" || getCurrencyShortName() == "bch_slp"))
		{
			const QRInfo qrinfoCashAddr = PostscriptRawGenerator::qrEncodeAddress(pubInfo.pubkey, pubInfo.accountName);			
			
			textWriter.addQR(qrinfoCashAddr, "pub cashaddr", "", 15.0, 200);					
		}
		
	}

	std::string ret = textWriter.getPs();

	ret += PostscriptLineGenerator::lineBetween(margin, 418, 540, 418);
		
	ret += "Text_font setfont\n";

	return ret;
}







std::string PostscriptGenerator::hierPriv(const std::string& privIndex, 
	const HierarcialInfo& hierarcialInfo,
	const std::string& pubAddr) const
{
	PSTextWriter textWriter(true, margin);

	Language lang = LANG_EN;

	textWriter.writeLn((lang == LANG_EN) ? "Raw private index" : "Privater Schlüssel");

	const std::string privIndexStart = privIndex.substr(0, 32);
	const std::string privIndexEnd = privIndex.substr(32);

	textWriter.writeLn(privIndexStart);
	textWriter.writeLn(privIndexEnd);

	textWriter.addNewline(2);

	textWriter.addPs("Big_font setfont\n");
	textWriter.writeLn("", (lang == LANG_EN) ? "CONFIDENTIAL" : "GEHEIMNIS");	
	textWriter.addPs("Text_font setfont\n");

	textWriter.addNewline(1);

	textWriter.writeLn("Should be combined with the private key corresponding to signature");
	textWriter.writeLn(hierarcialInfo.signature);
	textWriter.addNewline(1);





	textWriter.writeLn("Command:");	

	std::string theCommand;
	theCommand += "bitgen";

	const std::string currencyShortName = getCurrencyShortName();
	
	if(currencyShortName != "btc")
	{
		theCommand += "-";
		theCommand += currencyShortName;
	}
		
	theCommand += " genPriv root";
	theCommand += currencyShortName;
	theCommand += "_";
	theCommand += hierarcialInfo.signature;
	theCommand += "_priv.privroot ";

	textWriter.writeLn(theCommand);	

	std::string theFilename("num_" + intToString(info.getNewNum()) + "_" + pubAddr.substr(0, 8) + ".privindex");

	textWriter.writeLn(theFilename);	
	textWriter.addNewline(1);





	const std::string pubXstart = hierarcialInfo.pubX.substr(0, 32);
	const std::string pubXend = hierarcialInfo.pubX.substr(32);

	//textWriter.writeLn((lang == LANG_EN) ? "PubX" : "Privater Schlüssel");	
	textWriter.writeLn((lang == LANG_EN) ? "PubX" : "Privater Schlüssel", pubXstart);	
	textWriter.writeLn("", pubXend);	

	const std::string pubYstart = hierarcialInfo.pubY.substr(0, 32);
	const std::string pubYend = hierarcialInfo.pubY.substr(32);

	textWriter.writeLn("PubY", pubYstart);	
	textWriter.writeLn("", pubYend);	
	
	//textWriter.writeLn(pubYstart);	
	//textWriter.writeLn(pubYend);	
		
	const std::string ret = textWriter.getPs();

	return ret;	 
}





std::string PostscriptGenerator::addExtraTextLine(const int yPos, const std::string& thisPart, const int extraXindent) const
{
	std::string ret;
	
	ret += PostscriptRawGenerator::textAt(margin + extraXindent, yPos, thisPart);

	return ret;
}


/*
std::list<std::string> PostscriptGenerator::splitLongLineOnWordBoundry(const std::string& str)
{
	//std::cout << "Splitting line: " << str << std::endl;
	
	std::list<std::string> ret;
	std::string thisPart;
	for(int i = 0 ; i < str.size() ; i++)
	{
		const char c = str[i];
		thisPart += c;
		
		if(c == ' ' && thisPart.size() > 50)
		{
			ret.push_back(thisPart);
			thisPart = "";
		}
	}
	
	if(!thisPart.empty())
	{
		ret.push_back(thisPart);		
	}
	
	return ret;
}
*/


std::list<std::string> PostscriptGenerator::splitLongLine(const std::string& str, const bool splitWordBoundry, const int maxLineSize)
{
	//std::cout << "Splitting line: " << str << std::endl;
	
	std::list<std::string> ret;
	std::string thisPart;
	for(int i = 0 ; i < str.size() ; i++)
	{
		const char c = str[i];
		thisPart += c;
		
		if(splitWordBoundry)
		{
			if(c == ' ' && thisPart.size() > 50)
			{
				ret.push_back(thisPart);
				thisPart = "";
			}			
		}
		else
		{
			if(thisPart.size() >= maxLineSize)
			{
				ret.push_back(thisPart);
				thisPart = "";
			}
		}
	}
	
	if(!thisPart.empty())
	{
		ret.push_back(thisPart);		
	}
	
	return ret;
}



/*
std::string PostscriptGenerator::privateKeyPart(const PrivatePostscriptInfo& privInfo) const
{
	std::string ret;

	std::string privText;
	privText +=  ((lang == LANG_EN) ? "Private key" : "Privater Schlüssel");
	privText += " (WIF, Wallet Import Format)";
	
	ret += textAt(270,       350, privText);

	ret += textAt(margin,       390, privInfo.privkey);

	
	std::string genText;
	genText += (lang == LANG_EN) ? "Generation method" : "Erzeugungsverfahren";
	genText += ": ";
	genText += privInfo.generationMethod;
	
	ret += textAt(270,       220, genText);
	
	//if(getCurrencyShortName() != "xmr" && getCurrencyShortName() != "zec")
	//{	
	//	ret += textAt(270,       200, (privInfo.compressed) ? "Compressed format" : "Uncompressed format");
	//}

	int yPos(90);
	for(std::list<ExtraInfo>::const_iterator it = privInfo.generationExtra.begin() ;
		it != privInfo.generationExtra.end() ;
		++it)
	{
		const ExtraInfo& thisExtra = (*it);
		const std::string& thisPart = thisExtra.value;
		std::list<std::string> smallParts = splitLongLine(thisPart);
		
		for(std::list<std::string>::const_iterator jt = smallParts.begin() ;
			jt != smallParts.end() ;
			++jt)
		{
			const std::string& smallPart = (*jt);
			ret += addExtraTextLine(yPos, smallPart);
			yPos -= 20;
		}
	}




	ret += "Big_font setfont\n";


	ret += textAt(270,       300, (lang == LANG_EN) ? "SECRET" : "GEHEIMNIS");

	ret += textAt(270,       250, "Decrypt & sign");
	
	ret += "Text_font setfont\n";


	const std::string myEncStr = createPrivateEncString(privInfo);
	const QRInfo privinfo = QRInfo::qrencode((const unsigned char*)myEncStr.c_str(), myEncStr.size());
	ret += qrCode(privinfo, margin, 210, 20.0);

	return ret;
}
*/

std::list<std::string> PostscriptGenerator::splitMnemonic(const std::string& mnemonic, const int spacesPerPart)
{
	std::list<std::string> ret;

	int numSpace(0);
	int i(0);
	int lastPos(0);
	for( ; i < mnemonic.size() ; i++)
	{
		if(mnemonic[i] == ' ')
		{
			numSpace++;
			if(numSpace % spacesPerPart == 0)
			{
				const std::string part = mnemonic.substr(lastPos, i - lastPos);
				ret.push_back(part);
				lastPos = i + 1;				
			}
			//else if(numSpace == 16)
			//{
			//	break;
			//}

		}
	}
	
	const std::string lastPart = mnemonic.substr(lastPos, i - lastPos);
	ret.push_back(lastPart);

	return ret;
}


std::string PostscriptRawGenerator::generateQRPostscriptWithText(const QRInfo& qrinfo, 
											const std::string& lowerStr, const std::string& upperStr,
											const int x, const int y, const double scale) 
{		
	std::string ret;
	
	ret += PostscriptRawGenerator::qrCode(qrinfo, x, y, scale);
	ret += "Small_font setfont\n";
	ret += PostscriptRawGenerator::textAt(x, y - 10, lowerStr);
	ret += "Tiny_font setfont\n";	
	ret += PostscriptRawGenerator::textAt(x, y + 160, upperStr);
	
	return ret;
}



std::string PostscriptRawGenerator::generateQRPostscript(const QRInfo& qrinfo, const std::string& txtStr, 
											const int x, const int y, const double scale) 
{		
	std::string ret;
	
	ret += PostscriptRawGenerator::qrCode(qrinfo, x, y, scale);
	ret += "Small_font setfont\n";
	ret += PostscriptRawGenerator::textAt(x, y - 10, txtStr);
	
	return ret;
}


std::string PostscriptRawGenerator::generateQRPostscript(const std::string& qrStr, const std::string& txtStr, 
											const int x, const int y, const double scale) 
{		
	const QRInfo qrinfo = QRInfo::qrEncodeWif(qrStr);		
	
	return generateQRPostscript(qrinfo, txtStr, x, y, scale) ;
}





std::string PostscriptGenerator::privateMnemonic(const std::string mnemonic, const std::string& generationMethod,
	const std::list<ExtraInfo>& generationExtra, const std::string& xpriv) const
{
	PSTextWriter textWriter(true, margin);

	const std::pair<std::string, std::string> privParts = splitInTwoParts(xpriv);


	textWriter.writeLn(privParts.first);
	textWriter.writeLn(privParts.second);
	textWriter.addNewline(1);

	textWriter.writeTab(6, (lang == LANG_EN) ? "Private mnemonic" : "Privater mnemonic");
	textWriter.addNewline(2);

	textWriter.writeTab(6, (lang == LANG_EN) ? "SECRET" : "GEHEIMNIS");
	textWriter.addNewline(2);

	textWriter.writeTab(6, (lang == LANG_EN) ? "Spend" : "Verbringen");

	textWriter.addNewline(3);

	const bool hasXPriv = !xpriv.empty();

	if(hasXPriv)
	{
		const QRInfo qrinfo = QRInfo::qrEncodeWif(xpriv);		
		
		textWriter.addQR(qrinfo, "xpriv", "", 17.0);		
	}


	const QRInfo qrinfo = QRInfo::qrEncodeWif(mnemonic);		

	textWriter.addQR(qrinfo, "mnemonic", "", 17.0, true); //200);

	textWriter.addPs("Text_font setfont\n");

	textWriter.addNewline(2);

	textWriter.writeLn("Method:", generationMethod);

	const std::string mnemonicText = ( (addrType == ADDRTYPE_ELECTRUM) ? "EMnemonic:" : "Mnemonic BIP39 seed:");

	textWriter.writeLn(mnemonicText);

	std::list<std::string> mnemonicParts = splitMnemonic(mnemonic, 8);

	int wordNum(1);
	const int tabSize(0);

	for(const std::string& part : mnemonicParts)
	{
		const std::list<std::string> words = splitMnemonic(part, 1);	//Split into words

		int i(0);
		for(const std::string& word : words)
		{	
			//std::cout << "Writing word: " << word << std::endl;
			textWriter.writeTab(i, word);			
			textWriter.addPs("Tiny_font setfont\n");
			textWriter.writeTab(i, intToString(wordNum), 12);
			textWriter.addPs("Text_font setfont\n");
			
			i++;
			wordNum++;
		}

		textWriter.writeLn("");		
	}



	for(std::list<ExtraInfo>::const_iterator it = generationExtra.begin() ;
		it != generationExtra.end() ;
		++it)
	{
		const ExtraInfo& extraInfo = (*it);
		const std::string totTitle(extraInfo.title + ":");
		
		textWriter.writeLn(totTitle);
		
		const std::string& value = extraInfo.value;
		std::list<std::string> smallParts = splitLongLine(value);
		
		for(std::list<std::string>::const_iterator jt = smallParts.begin() ;
			jt != smallParts.end() ;
			++jt)
		{
			const std::string& smallPart = (*jt);
			textWriter.writeLn(smallPart);			
		}
	}

	//if(hasXPriv)
	//{
	//	ret += "Titel_font setfont\n";
	//}
	//else
	//{
	//	ret += "Big_font setfont\n";		
	//}
	//ret += "Text_font setfont\n";

	const std::string ret = textWriter.getPs();
	
	return ret;
}










/*
	std::string text = (lang == LANG_EN) ? "Private key" : "Privater Schlüssel";
	
	if(!extended)
	{
		text += " (WIF, Wallet Import Format)";
	}


	ret += textAt(270, 350, text);
*/

/*
	if(isExtended())
	{
		const std::pair<std::string, std::string> privkeyParts = splitInTwoParts(privInfo.privkey);

		const std::string privLeft = privkeyParts.first;
		const std::string privRight = privkeyParts.second;
	
		textWriter.writeLn(privLeft);
		textWriter.writeLn(privRight);		
	}
	else if(privInfo.inhibitFirstPrivkeyLine)
	{
		//Nothing
	}
*/	


std::string PostscriptGenerator::privatePart(const PrivatePostscriptInfo& privInfo, const bool useBigInfoFont) const
{	
	PSTextWriter textWriter(true, margin); //, 100);

	//textWriter.addNewline(1);

	if(privInfo.isPad)	
	{
		const std::string padTxt = "PAD " + intToString(privInfo.padNum) + "/" + intToString(privInfo.totalPads);
		textWriter.writeSec2("", padTxt);
	}
		
	textWriter.addNewline(3);


	const bool hasMnemonic = !privInfo.mnemonic.empty();


		
		
	std::list<QRData> qrCodes;
	
	if(isExtended())
	{
		const QRInfo qrinfo = QRInfo::qrEncodeWif(privInfo.privkey);
		const QRData qrData(qrinfo, "xpriv", "");
		qrCodes.push_back(qrData);
	

		//textWriter.addQR(qrData, 15.0);
		//textWriter.addQR(qrinfo, "xpriv", "", 15.0);
		
		if(hasMnemonic)
		{
			const std::string mnemonicText = ( (addrType == ADDRTYPE_ELECTRUM) ? "Electrum mnemonic" : "mnemonic");
						
			const QRInfo qrinfo = QRInfo::qrEncodeWif(privInfo.mnemonic);
			//qrCodes.push_back(qrinfo);
			const QRData qrData2(qrinfo, mnemonicText, "");

			//textWriter.addQR(qrData2, /*qrinfo, mnemonicText, "",*/ 15.0, true);
		}
	}
	else
	{		
		if(privInfo.hasBip39Mnemonic())
		{			
			const QRInfo qrinfo = QRInfo::qrEncodeWif(privInfo.mnemonic);				
			const QRData qrData(qrinfo, "BIP 39 mnemonic", "");
			qrCodes.push_back(qrData);
			
			//textWriter.addQR(qrData, /*qrinfo, "BIP 39 mnemonic", "",*/ 15.0);					
		}
		
		if(privInfo.isPad)
		{
		}
		else
		{
			const QRInfo qrinfo = QRInfo::qrEncodeWif(privInfo.privkey);				
			const QRData qrData(qrinfo, privInfo.privkeyName, "");
			qrCodes.push_back(qrData);												
		}		
	}


	const int numQr = qrCodes.size();
	
	if(useBigInfoFont)
	{
		textWriter.addPs("Big_font setfont\n");
	}
	else
	{
		textWriter.addPs("Titel_font setfont\n");
	}

	if(numQr == 0 || numQr == 1)
	{
		textWriter.writeSec2((lang == LANG_EN) ? "SECRET" : "GEHEIMNIS", "");
		textWriter.addNewline(2);

		textWriter.writeSec2((lang == LANG_EN) ? "Spend" : "Verbringen", "");

	}
	else
	{
		textWriter.writeSec2("", (lang == LANG_EN) ? "SECRET" : "GEHEIMNIS");
		textWriter.addNewline(2);
		textWriter.writeSec2("", (lang == LANG_EN) ? "Spend" : "Verbringen");

	}
	
	textWriter.addPs("Titel_font setfont\n");

	textWriter.addNewline(2);
	
	
	int i(0);
	for(const QRData& qd : qrCodes)
	{
		textWriter.addQR(qd, 15.0, (i == 0 ? false : true));
		i++;
	}
	
	
	
	textWriter.addPs("Text_font setfont\n");
		
	textWriter.addNewline(2);
		
		
	if(privInfo.isPad)	
	{
		const std::pair<std::string, std::string> parts = splitInTwoParts(privInfo.privkey);

		textWriter.writeLn(parts.first);
		textWriter.writeLn(parts.second);

		textWriter.addPs("Small_font setfont\n");
		textWriter.writeLn(privInfo.privkey);

		textWriter.addPs("Text_font setfont\n");			
	}
	else
	{
		if(privInfo.privkey.size() > 50 && privInfo.privkey.size() < 100)		//Can fit on one row with small font
		{		
			textWriter.addPs("Small_font setfont\n");
			textWriter.writeLn(privInfo.privkeyName + ":", privInfo.privkey);	
			textWriter.addPs("Text_font setfont\n");					
		}
		else
		{
			const std::list<std::string> privateParts = splitInNParts(privInfo.privkey, 56);
			
			bool isFirst(true);
			for(const std::string privPart : privateParts)
			{					
				if(isFirst)
				{
					isFirst = false;
					textWriter.writeLn(privInfo.privkeyName + ":", privPart);					
				}
				else
				{
					textWriter.writeLn("", privPart);				
				}
			}
		}
	}

	textWriter.addPs("Small_font setfont\n");


	std::string printText;
	printText += ((lang == LANG_EN) ? "Method" : "Erzeugungsverfahren");
	printText += ":";

	const BigInt<256> zero;

	const std::string hashBin = staticXorHash(privInfo.staticXor);
	const std::string hashStr = hexString(hashBin);

	const std::string staticExtra = ((privInfo.staticXor != zero) ? " with static XOR (" + hashStr + ")": "");
	

	textWriter.writeLn(printText, privInfo.generationMethod + staticExtra);

	textWriter.addPs("Text_font setfont\n");					


	if(/*!isExtended() &&*/ !privInfo.mnemonic.empty())
	{
		
		const std::string mnemonicText = ( (addrType == ADDRTYPE_ELECTRUM) ? "Electrum mnemonic:" : "Mnemonic:");		

		textWriter.writeLn(mnemonicText);
	
		const std::list<std::string> mnemonicParts = splitMnemonic(privInfo.mnemonic, 8);

		int wordNum(1);

		for(const std::string& part : mnemonicParts)
		{
			int i(0);
			const std::list<std::string> words = splitMnemonic(part, 1);
			
			for(const std::string& word : words)
			{
				textWriter.writeTab(i, word);
			
				textWriter.addPs("Tiny_font setfont\n");
				textWriter.writeTab(i, intToString(wordNum) /*word*/, 12);
				
				textWriter.addPs("Text_font setfont\n");
			
				i++;
				wordNum++;
			}		
			textWriter.writeLn("");			
		}
	}




	for(std::list<ExtraInfo>::const_iterator it = privInfo.generationExtra.begin() ;
		it != privInfo.generationExtra.end() ;
		++it)
	{
		const ExtraInfo& extraInfo = (*it);
		const std::string totTitle(extraInfo.title + ":");
		const std::string& value = extraInfo.value;

		if(value.size() > 50 && value.size() < 100)		//Can fit on one row with small font
		{		
			//textWriter.writeLn("");						
			textWriter.addPs("Small_font setfont\n");
			textWriter.writeLn(totTitle, value);	
			textWriter.addPs("Text_font setfont\n");					
		}
		else //Split on multiple lines if needed 
		{
		textWriter.write(totTitle, "");
		//std::cout << "TITLE: " << totTitle << std::endl;
			
		std::list<std::string> smallParts = splitLongLine(value, extraInfo.splitWords, 50);
		
		for(std::list<std::string>::const_iterator jt = smallParts.begin() ;
			jt != smallParts.end() ;
			++jt)
		{
			const std::string& smallPart = (*jt);
			textWriter.write("", smallPart);
			textWriter.writeLn("");
		}
		}
	}




	if(hasMnemonic)
	{
		textWriter.addPs("Titel_font setfont\n");
	}
	else
	{
		textWriter.addPs("Big_font setfont\n");
	}





	std::string ret = textWriter.getPs();

	return ret;
}




std::string PostscriptGenerator::staticXorInfo(const std::string& thetime, const BigInt<256>& staticXor) const
{
	const int columnWidth(100);
	
	PSTextWriter textWriter(false, margin, columnWidth);
	
	const std::string hash = staticXorHash(staticXor);

	textWriter.addPs("Text_font setfont\n");

	const std::string header = "Static XOR " + hexString(hash);

	textWriter.writeLn(header);

	const std::string bitgenVersion = applicationName + " " + versionNumber;


	textWriter.writeLn("Generated by:", bitgenVersion);
	textWriter.writeLn("Generate date:", thetime);
	textWriter.writeLn("Hash: ", hexString(hash));

	textWriter.addNewline(5);
	textWriter.addPs("Big_font setfont\n");

	textWriter.writeLn("SECRET");
	textWriter.addNewline(1);
	textWriter.writeLn("STATIC XOR");
	textWriter.addNewline(5);

	textWriter.addPs("Text_font setfont\n");

	textWriter.writeLn("This static xor key is used to generate split wallets.");
	textWriter.writeLn("To be able to recreate a private key the split xor key MUST BE PRESENT.");
	textWriter.writeLn("It is very important that this private xor key is securely stored");
	
	textWriter.addNewline(2);

	const std::string xorStrFull = hexString(staticXor.getBinaryLong());
	const std::pair<std::string, std::string> xorParts = splitInTwoParts(xorStrFull);

	textWriter.writeLn(xorParts.first);
	textWriter.writeLn(xorParts.second);
	
	const std::string ret = textWriter.getPs();
	return ret;
}




std::string PostscriptGenerator::generateStaticXor(
	const std::string& thetime, const BigInt<256>& staticXor) const
{
	const int margin(70);
	
	std::string ret;

	ret += PostscriptRawGenerator::intro();
	
	ret += PostscriptRawGenerator::beginPage();
	
	
	ret += staticXorInfo(thetime, staticXor);
	
	ret += PostscriptRawGenerator::endPage();
	
	ret += "%%EOF\n";
	return ret;
}




std::string PostscriptGenerator::generatePrivateRoot(const GenerationInfo& info,
	const HierarcialInfo& hierarcialInfo,
	const std::string& thetime, const std::string& label,
	const std::string& privIndex) const
{
	const int margin(70);
	
	std::string ret;

	ret += PostscriptRawGenerator::intro();
	
	ret += PostscriptRawGenerator::beginPage();
	
	const AddressGenerationInfo addrInfo("dummyPubVal", "", "", info);
	
	const PublicPostscriptInfo pubInfo(addrInfo);
	
	ret += publicRoot(hierarcialInfo, pubInfo, label);
	
	ret += PostscriptLineGenerator::lineBetween(margin, 418, 540, 418);
	
	ret += privateRoot(privIndex, hierarcialInfo.signature);
	
	ret += PostscriptRawGenerator::endPage();
	
	ret += "%%EOF\n";
	return ret;
}




std::string PostscriptGenerator::generatePublicRoot(const HierarcialInfo& hierarcialInfo,
	const PublicPostscriptInfo& pubInfo, const std::string& label) const
{
	const int margin(70);
	
	std::string ret;

	ret += PostscriptRawGenerator::intro();
	
	ret += PostscriptRawGenerator::beginPage();

	ret += publicRoot(hierarcialInfo, pubInfo, label);
	
	ret += PostscriptLineGenerator::lineBetween(margin, 418, 540, 418);
	
	ret += PostscriptRawGenerator::endPage();
	
	ret += "%%EOF\n";
	return ret;
}




std::string PostscriptGenerator::postscriptRequestPaymentPage(
							const PublicPostscriptInfo& pubInfo,
							const std::string& amount,
							const std::string& message,
							const std::string& thetime) const
{		
	std::string ret;

	ret += PostscriptRawGenerator::intro();
	
	ret += PostscriptRawGenerator::beginPage();
		
	ret += generatePaymentRequest(pubInfo, amount, message, thetime);
	
	ret += PostscriptRawGenerator::endPage();
	
	ret += "%%EOF\n";

	return ret;
}






std::string PostscriptGenerator::generateInvoice(const PublicPostscriptInfo& pubInfo,
								const std::string amount, const std::string& message,
								const std::string& thetime,
								const std::string paymentId) const
{		
	std::string ret;

	ret += PostscriptRawGenerator::intro();
	
	ret += PostscriptRawGenerator::beginPage();
		
	ret += generatePaymentRequest(pubInfo, amount, message, thetime, paymentId);
	
	ret += PostscriptRawGenerator::endPage();
	
	ret += "%%EOF\n";

	return ret;
}



std::string PostscriptGenerator::privateHierarcialPrivate(
							const std::string& privIndex,
							const PublicPostscriptInfo& pubInfo,
							const HierarcialInfo& hierarcialInfo
							) const
{		
	std::string ret;

	ret += PostscriptRawGenerator::intro();
	
	ret += PostscriptRawGenerator::beginPage();
		
	ret += publicPart(pubInfo, true, true);
	
	ret += hierPriv(privIndex, hierarcialInfo, pubInfo.pubkey);
	
	ret += PostscriptRawGenerator::endPage();
	
	ret += "%%EOF\n";

	return ret;
}



std::string PostscriptGenerator::postscriptPagePublic(const PublicPostscriptInfo& pubInfo) const
{	
	std::string ret;

	ret += PostscriptRawGenerator::intro();
	
	ret += PostscriptRawGenerator::beginPage();
		
	ret += publicPart(pubInfo, true, true);
	
	ret += PostscriptRawGenerator::endPage();
	
	ret += "%%EOF\n";

	return ret;
}



std::string PostscriptGenerator::postscriptPagePublicExtended(const PublicExtendedPostscriptInfo& pubInfo) const
{	
	std::string ret;

	ret += PostscriptRawGenerator::intro();
	
	ret += PostscriptRawGenerator::beginPage();
		
	ret += publicPartExtended(pubInfo, false);
	
	ret += PostscriptRawGenerator::endPage();
	
	ret += "%%EOF\n";

	return ret;
}


/*
std::string PostscriptGenerator::postscriptPagePublicKey(const PublicPostscriptInfo& pubInfo) const
{	
	std::string ret;

	ret += intro();
	
	ret += beginPage();
	
	ret += publicKeyPart(pubInfo);
	
	ret += endPage();
	
	ret += "%%EOF\n";

	return ret;
}
*/



/*
std::string PostscriptGenerator::pubPrivPairPostscriptPage(
							const bool isPrivate, 
							const PrivatePostscriptInfo& privInfo,
							const PublicPostscriptInfo& pubInfo
							) const
{	
	std::string ret;

	ret += intro();
	
	ret += beginPage();
	
	ret += publicKeyPart(pubInfo);
	
	//std::cout << "Will start with private part" << std::endl;
	if(isPrivate)
	{
		ret += privateKeyPart(privInfo);
	}

	ret += endPage();
	
	ret += "%%EOF\n";

	return ret;
}
*/



std::string PostscriptGenerator::generatePublicMnemonic(
	const std::string mnemonic, 
	const int numWords, 
	const std::string& accountName,	
	//const std::string& not_used_generationMethod,
	//const std::string& not_used_theTime, 
	//const std::list<ExtraInfo>& generationExtra,
	const std::string& xpriv,
	const ExtendedPublicKey& xpub,
	const std::string& derivedXpub) const
{	
	std::string ret;

	ret += PostscriptRawGenerator::intro();
	
	ret += PostscriptRawGenerator::beginPage();
	
	const PublicExtendedPostscriptInfo pubInfo(info.getNodeName(), xpub.toString(), info.getTime(), accountName);
	
	const bool useSmallDescription = !xpriv.empty();
	ret += publicMnemonic(pubInfo, useSmallDescription, numWords, derivedXpub);
	
	//ret += privateMnemonic(mnemonic, numWords, generationMethod, generationExtra, xpriv);

	ret += PostscriptRawGenerator::endPage();
	
	ret += "%%EOF\n";

	return ret;
}



std::string PostscriptGenerator::generatePrivateMnemonic(
	const std::string mnemonic, 
	const int numWords, 
	const std::string& accountName,	
	const std::string& xpriv,
	const ExtendedPublicKey& xpub,
	const std::string& derivedXpub) const
{	
	std::string ret;

	ret += PostscriptRawGenerator::intro();
	
	ret += PostscriptRawGenerator::beginPage();
	
	const PublicExtendedPostscriptInfo pubInfo(info.getNodeName(), xpub.toString(), info.getTime(), accountName);
	
	const bool useSmallDescription = !xpriv.empty();
	ret += publicMnemonic(pubInfo, useSmallDescription, numWords, derivedXpub);
	
	ret += privateMnemonic(mnemonic, info.getGenerator(), info.getGenerationExtra(), xpriv);

	ret += PostscriptRawGenerator::endPage();
	
	ret += "%%EOF\n";

	return ret;
}






std::string PostscriptGenerator::generatePrivate(
							const bool isPrivate, 
							const PrivatePostscriptInfo& privInfo,
							const PublicPostscriptInfo& pubInfo,
							const bool useBigInfoFont
							) const
{	
	std::string ret;

	ret += PostscriptRawGenerator::intro();
	
	ret += PostscriptRawGenerator::beginPage();
	
	const bool useSmallText = !privInfo.mnemonic.empty();
	ret += publicPart(pubInfo, useSmallText, useBigInfoFont);
	
	if(isPrivate)
	{
		ret += privatePart(privInfo, useBigInfoFont);
	}

	ret += PostscriptRawGenerator::endPage();
	
	ret += "%%EOF\n";

	return ret;
}




std::string PostscriptGenerator::generatePrivateExtended(
							const bool isPrivate, 
							const PrivatePostscriptInfo& privInfo,
							const PublicExtendedPostscriptInfo& pubInfo
							) const
{	
	std::string ret;

	ret += PostscriptRawGenerator::intro();
	
	ret += PostscriptRawGenerator::beginPage();
	
	//const bool useSmallText = !privInfo.mnemonic.empty();
	ret += publicPartExtended(pubInfo, true); //useSmallText);
	
	if(isPrivate)
	{
		ret += privatePart(privInfo, true);
	}

	ret += PostscriptRawGenerator::endPage();
	
	ret += "%%EOF\n";

	return ret;
}








