/**
 * bitcoin.cpp - Bitcoin address generator
 *
 * Copyright (C)2015 The bitgen developers <bitgen@openmailbox.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
 
/**
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE. 
*/
 

#include "bitgenVersion.h"
#include "currencySpecific.h"
#include "util.h"
#include "license.h"
#include "bech32.h"
#include "extraInfo.h"
#include "GenerationParameters.h"
#include "GenerationMethods.h"

#include "commandLine.h"
#include "bitgencli.h"

#include "bitgenConfig.h"

#include <iostream>

#include <time.h>

void printZcashzUsage()
{
	std::cout << "Usage: bitgen-zecz gen secretKey publicAddr" << std::endl;
	std::cout << std::endl;
}




void verifyPrivateKey(const std::string& privateKey)
{
	const std::pair<std::string, std::vector<uint8_t>> decoded = bech32::Decode(privateKey);

	const std::string hrp = decoded.first;
	
	const std::vector<uint8_t> data = decoded.second;

	//std::cout << "Private key hrp: " << hrp << std::endl;
	
	if(data.size() != 271)
	{
		throw std::string("Incorrect data size for private key");
	}

	if(hrp != "secret-extended-key-main")
	{
		std::cout << "Incorrect Human readable part" << std::endl;
		throw std::string("Incorrect human readable part");
	}
}


void verifyZcashzAddress(const std::string& addr)
{
	const std::pair<std::string, std::vector<uint8_t>> decoded = bech32::Decode(addr);

	const std::string hrp = decoded.first;
	
	const std::vector<uint8_t> data = decoded.second;

	//std::cout << "Address hrp: " << hrp << std::endl;
	
	if(data.size() != 69)
	{
		std::cout << "Incorrect data size for address: " << data.size() << std::endl;
		throw std::string("Incorrect data size for address");
	}

	if(hrp != "zs")
	{
		std::cout << "Incorrect Human readable part" << std::endl;
		throw std::string("Incorrect human readable part");
	}
}


void GenerationMethods::generateZcashz(const std::vector<std::string>& args, const std::string& thetime)
{
	if(args.size() < 3)
	{
		std::cout << std::endl;
		std::cout << "Usage: bitgen-zecz gen privateKey publicAddress" << std::endl;
		return;
	}
	
	std::cout << std::endl;
		
	const std::string privateKey(args[1]);
	const std::string addr(args[2]);
		
	verifyPrivateKey(privateKey);
	verifyZcashzAddress(addr);

	std::list<ExtraInfo> privateGenerationExtra;

	GenerationParameters params;

	//const ExtraInfo extraInfoPrivSpend("AAA", "VAL");
	//privateGenerationExtra.push_back(extraInfoPrivSpend);

	//const int newNum = GenerationMethods::getNextNumberIfConfigured(params);
	const int newNum = BitgenConfig::getConfiguredNum();


	//params.setAccountNameForNumber(true, newNum);

	const GenerationInfoBase infoBase(thetime, getNodeName(), applicationName, versionNumber);
	const GenerationInfoCore infoCore(infoBase, newNum, params);
	const GenerationInfo info(infoCore, privateGenerationExtra, "External");
	
	//info.setCashAddressRead("");
	//info.setCashAddress("");
	//info.setMnemonic("");
	//info.setIsPad(false);
	//info.setStaticXor(BigInt<256>());
	//info.setInhibitFirstPrivkeyLine(false);
	
	const AddressGenerationInfo allInfo(addr, privateKey, "Priv", info);

	GenerationMethods::saveAllFiles(allInfo);
		//addr, "", "",
		//privateKey, "External", 
		//privateGenerationExtra, thetime, params, "", newNum, false, BigInt<256>(), false);	

}




void doZcashzWork(int argc, char* argv[])
{			
	time_t t = time(0);
	const std::string thetime = removeNewline(asctime(localtime(&t)));
	
	std::vector<std::string> args;
	for(int i = 1 ; i < argc ; i++)
	{
		args.push_back(argv[i]);		
	}

	if(argc < 2)
	{
		std::cout << "For help, use \"bitgen-" << getCurrencyShortName() << " help\"" << std::endl;
		throw std::string("Missing argument");
		return;
	}

	const std::string firstArg(argv[1]);


	if(firstArg == "version" || firstArg == "-v")
	{
		printCompleteVersion();
	}
	else if(firstArg == "help")
	{
		printZcashzUsage();	
	}
	else if(firstArg == "gen")
	{
		GenerationMethods::generateZcashz(args, thetime);	
	}	
	else if(firstArg == "license")
	{
		printLicense();
	}
	else if(firstArg == "getnextnum")
	{
		std::cout << "Getting next num" << std::endl;
		getNextNumToUse();
	}
	else if(firstArg == "setnextnum")
	{		
		std::cout << "Setting next num" << std::endl;
		setNextNumToUse(args);
	}	
	else
	{
		std::cerr << "Unknown command: " << argv[1] << std::endl;
	}
	
}









int main(int argc, char* argv[])
{				
	try
	{
		if(argc > 1)
		{		
		const std::string firstArg(argv[1]);
		if(firstArg == "test")
		{
			//unitTestsLtc();
		
#ifndef NO_BRAINWALLET_GENERATION
		//brainwalletTests();
#endif	
			std::cout << std::endl;
			std::cout << "All tests performed successfully" << std::endl;	
			return 0;
		}
		}
		
		doZcashzWork(argc, argv);
	}
	catch(const std::string& err)
	{
		std::cerr << "Error: " << err << std::endl;
		return 1;
	}
	
	return 0;
}


	
