

#include "pubXmrAddress.h"

#include "keccakHash.h"

#include "base58.h"
#include "bigintutil.h"
#include "util.h"




XmrPubKey PubXmrAddress::getSpendKey() const
{
	XmrEllipticCurve ec;
	const XmrECPoint point = ec.recoverPoint(pubSpendVal);
	const XmrPubKey pubkey(point);
	return pubkey; //point;
}


XmrPubKey PubXmrAddress::getViewKey() const
{
	XmrEllipticCurve ec;
	const XmrECPoint point = ec.recoverPoint(pubViewVal);
	const XmrPubKey pubkey(point);
	return pubkey; //return point;
}






PubXmrAddress::PubXmrAddress(const BigInt<257>& inPubSpend, const BigInt<257>& inPubView, const bool inIsSubaddress) : 
	pubSpendVal(inPubSpend), pubViewVal(inPubView), isSubaddress(inIsSubaddress)
{
	//Todo: Check validity
}


PubXmrAddress::PubXmrAddress(const XmrPubKey& inPubSpend, const XmrPubKey& inPubView, const bool inIsSubaddress) : 
	pubSpendVal(inPubSpend.xyVal()), pubViewVal(inPubView.xyVal()), isSubaddress(inIsSubaddress)
{
	//Todo: Check validity
}


PubXmrAddress::PubXmrAddress(const PubXmrAddress& other, const BigInt<65>& inPaymentId) : 
	isSubaddress(other.isSubaddress),
	pubSpendVal(other.pubSpendVal),
	pubViewVal(other.pubViewVal),
	hasPaymentId(true),
	paymentId(inPaymentId)
{	
}



PubXmrAddress::PubXmrAddress(const PubXmrAddress& other, const bool dummyNoPaymentId) : 
	isSubaddress(other.isSubaddress),
	pubSpendVal(other.pubSpendVal),
	pubViewVal(other.pubViewVal),
	hasPaymentId(false)
{	
}



BigInt<800> split11ByteBlocks(const std::string& pubAddr)
{	
	BigInt<800> totResult;

	//Split in 11 byte blocks
	
	for(int i = 0 ; ; i++)
	{
		if(i*11 > pubAddr.size())
		{
			break;
		}
		
		const std::string first = pubAddr.substr(i*11, 11);
		//const std::string firstRev = doreverseOrder(first);
			
		const BigInt<800> result = Base58Decode<800>(first);

		if(first.size() < 11)
		{
			totResult <<= 40;
			totResult += result;			
		}
		else
		{
			totResult <<= 64;
			totResult += result;
		}		
	}


	return totResult;
}




PubXmrAddress::PubXmrAddress(const std::string& pubAddr)
{	
	if(pubAddr.size() != 95 && pubAddr.size() != 106)
	{
		std::string errStr("Xmr address must be exactly 95 or 106 characters, got: ");
		errStr += intToString(pubAddr.size());
		throw errStr; //std::string("Xmr address must be exactly 95 or 106 characters");		
	}

	const BigInt<800> totResult = split11ByteBlocks(pubAddr);

	const bool isIntegratedAddress = (pubAddr.size() == 106);

	const int paymentIdSize = ((pubAddr.size() == 95) ? 0 : 64);

	const BigInt<800> network = totResult >> (544 + paymentIdSize);
	const int netVal = network.intVal();

	if(isIntegratedAddress)
	{
		if(netVal != 19 && netVal != 42)
		{
			std::cout << "Got netval: " << netVal << std::endl;
			throw std::string("Error, only xmr main chain supported");
		}
	}
	else //if(pubAddr.size() == 95)
	{
		if(netVal != 18 && netVal != 42)
		{
			throw std::string("Error, only xmr main chain supported");
		}		
	}

	isSubaddress = (netVal == 42) ? true : false;

	const BigInt<800> pubSpend = (totResult >> (288 + paymentIdSize)) & 256;
	const BigInt<257> myPubSpendVal = reverseIntBytes<800>(pubSpend);

	pubSpendVal = myPubSpendVal;
		
	const BigInt<800> pubView = (totResult >> (32 + paymentIdSize)) & 256;
	const BigInt<257> myPubViewVal = reverseIntBytes<800>(pubView);	
	pubViewVal = myPubViewVal;

	if(isIntegratedAddress) //pubAddr.size() == 106)
	{
		const BigInt<800> myPaymentId = (totResult >> (32)) & 64;
		paymentId = BigInt<65>::fromBytes(myPaymentId.getBinaryLong());
		hasPaymentId = true;
	}

	const BigInt<800> checksum = (totResult >> 0) & 32;

	const BigInt<800> forChecksumComputation = (totResult >> 32) & (520 + paymentIdSize);

	const std::string binBuf = forChecksumComputation.getBinary();


	if(pubAddr.size() == 95)
	{
		if(binBuf.size() != 65)
		{
			throw std::string("Internal error, buffer for checksum computation should be 65 bytes");
		}
	}
	else
	{
		if(binBuf.size() != 73)
		{
			throw std::string("Internal error, buffer for checksum computation should be 73 bytes");
		}
	}

	const std::string r = computeKeccak1600(binBuf);

	const std::string computedChecksum = r.substr(0, 4);

	const BigInt<800> computedChecksumInt(BigInt<800>::fromBytes(computedChecksum));

	if(checksum != computedChecksumInt)
	{
		std::cout << "Given checksum   : " << checksum << std::endl;
		std::cout << "Computed checksum: " << computedChecksumInt << std::endl;
		
		throw std::string("Checksum error");
	}
}




std::string prefixOnes(std::string str, const int num)
{
	while(str.size() < num)
	{
		str = std::string("1") + str;
	}
	
	return str;
}


std::string xmrBase58(const std::string& buf, bool hasPaymentId)
{
	std::string result;
	
	const int loopBound = (hasPaymentId ? 10 : 9);
	
	for(int i = 0 ; i < loopBound ; i++)
	{
		//std::cout << "i: " << i << std::endl;

		if(i == (loopBound - 1))
		{
			const std::string part = buf.substr(i * 8, 5);
			const BigInt<41> val(BigInt<41>::fromBytes(part));
	
			//std::cout << "To be encoded: " << val << std::endl;
	
			const std::string encodedPart = Base58Encode<41>(val);	
			const std::string encoded = prefixOnes(encodedPart, 7);
			//std::cout << "encoded: " << encoded << std::endl;
			result += encoded;	
			
		}
		else
		{
			const std::string part = buf.substr(i * 8, 8);
			const BigInt<65> val(BigInt<65>::fromBytes(part));
	
			//std::cout << "To be encoded: " << val << std::endl;
	
			const std::string encodedPart = Base58Encode<65>(val);				
			const std::string encoded = prefixOnes(encodedPart, 11);
			//std::cout << "encoded: " << encoded << std::endl;


			result += encoded;	
		}
	
	}
	
	
	
	return result;
}



int getNetworkNum(const bool isSubaddress, const bool hasPaymentId)
{
		if(hasPaymentId)
		{
			return 19;
		}
		else
		{
			return isSubaddress ? 42 : 18;
		}
}



std::string PubXmrAddress::getPaymentval() const
{
	//std::string buf;

	if(hasPaymentId)
	{
		const std::string paymentIdRaw = paymentId.getBinaryLong();			
		const std::string trimedPaymentval = paymentIdRaw.substr(1, 8);	//Last 8 bytes
		
		return trimedPaymentval;
	}
	else
	{
		return "";
	}
}




std::string PubXmrAddress::toString() const
{
	const int networkNum = getNetworkNum(isSubaddress, hasPaymentId); //isSubaddress ? 42 : 18;
	
	const BigInt<8> network(networkNum);
	
	std::string result;
	
	const std::string netStr = network.getBinary();
	const std::string netStrByte = netStr.substr(netStr.size() - 1);
		
	const std::string spendValRaw = pubSpendVal.getBinaryLong();	
	//std::cout << "spendValRaw: " << hexString(spendValRaw) << std::endl;
	
	const std::string spendVal = reverseStr(spendValRaw);	
	//std::cout << "spendVal     : " << hexString(spendVal) << std::endl;
	
	const std::string trimedSval = spendVal.substr(0, 32); //spendVal.size() - 32);
	//std::cout << "trimedSval   : " << hexString(trimedSval) << std::endl;
	
	const std::string viewValRaw = pubViewVal.getBinaryLong();	
	const std::string viewVal = reverseStr(viewValRaw);	
	//std::cout << "viewVal        : " << hexString(viewVal) << std::endl;


	const std::string trimedVval = viewVal.substr(0, 32); //viewVal.size() - 32);
	//std::cout << "trimedVval     : " << hexString(trimedVval) << std::endl;

	const std::string trimedPaymentval = getPaymentval();
		 
	const std::string buf = netStrByte + trimedSval + trimedVval + trimedPaymentval;

	if(hasPaymentId)
	{
		if(buf.size() != 73)
		{
			throw std::string("Internal error, the buffer for checksum computation should be exactly 73 bytes");
		}
	}
	else
	{
		if(buf.size() != 65)
		{
			throw std::string("Internal error, the buffer for checksum computation should be exactly 65 bytes");
		}
	}

	const std::string keccakHash = computeKeccak1600(buf);
	const std::string computedChecksum = keccakHash.substr(0, 4);
	const std::string totalBuf = buf + computedChecksum;
	const std::string finalAddress = xmrBase58(totalBuf, hasPaymentId);

	return finalAddress;
}




