/**
 * privKey.cpp- Bitcoin poscript generator
 *
 * Copyright (C)2015 The bitgen developers <bitgen@openmailbox.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
 
 
#include "xmrPrivKey.h"

#include "pubXmrAddress.h"	//Todo: Remove

#include "util.h"
#include "keccakHash.h"

std::string XmrPrivKey::toBinaryXmrString() const
{
	const std::string binBuf = key.getBinaryLong();
	return reverseStr(binBuf);
}




//XmrPrivKey 
XmrPrivViewKey XmrPrivSpendKey::privateSpendToView() const
{
	if(!isSpendKey)
	{
		std::cout << "Error, only spend key can be converted to view key" << std::endl;
		throw std::string("Error, only spend key can be converted to view key");
	}
	
	const std::string privkeyRaw = toBinaryXmrString();

	const std::string hashViewKey = computeKeccak1600(privkeyRaw);
	
	const std::string hashViewKeyReversed = reverseStr(hashViewKey);

	const BigInt<256> hashAsInt(BigInt<256>::fromBytes(hashViewKeyReversed));

	BigInt<256> finalVal = XmrEllipticCurve::getEdwards25519Max(hashAsInt);

	const std::string finalValStr = finalVal.getBinaryLong();

	const std::string finalValStrRev = reverseStr(finalValStr);

	BigInt<256> privateViewKey(BigInt<256>::fromBytes(finalValStr));

	const XmrPrivViewKey myViewKey(privateViewKey);

	return myViewKey;
}




std::string XmrPrivKey::toString() const
{
	return hexString(reverseStr(key.getBinaryFull()));
}



std::ostream& operator<<(std::ostream& str, const XmrPrivKey& k)
{
	str << hexString(reverseStr(k.key.getBinaryFull()));
	//str << hexString(reverseIntBytes<256>(k.key).getBinaryLong());
	
	return str;
}





XmrPubKey XmrPrivKey::getPubkey(const XmrEllipticCurve& ec) const
{
	const XmrECPoint pubpoint = point(ec);		
	const XmrPubKey pubkey(pubpoint);
	return pubkey;
}


XmrECPoint XmrPrivKey::point(const XmrEllipticCurve& ec) const
{			
	const BigInt<256> privKeyFinal = XmrEllipticCurve::getEdwards25519Max(key);
	
	const BigInt<2048> privKeyRoot(BigInt<2048>::asPositive(privKeyFinal));
			
	const XmrECPoint PublicKey = ec.multiplyGenPoint(privKeyRoot);
	
	return PublicKey;
}



//Standard address for private spend key
PubXmrAddress XmrPrivSpendKey::getStandardAddress() const
{	
	if(!isSpendKey)
	{
		throw std::string("Internal error, getStandardAddressView should only be used for private spend key");
	}
	
	XmrEllipticCurve ec;

	const XmrPubKey pubKeySpend = getPubkey(ec);

	const XmrPrivViewKey privViewKey = privateSpendToView();
	const XmrPubKey pubKeyView = privViewKey.getPubkey(ec);

	const PubXmrAddress pubXmrAddress(pubKeySpend, pubKeyView);
	
	return pubXmrAddress;
}


//Standard address for private view key
PubXmrAddress XmrPrivViewKey::getStandardAddressView(const XmrPubKey& pubKeySpend) const
{	
	if(isSpendKey)
	{
		throw std::string("Internal error, getStandardAddressView should only be used for private view key");
	}
	
	XmrEllipticCurve ec;

	const XmrPubKey pubKeyView = getPubkey(ec);

	const PubXmrAddress pubXmrAddress(pubKeySpend, pubKeyView);
	
	return pubXmrAddress;
}



//Subaddress for private view key
PubXmrAddress XmrPrivViewKey::getSubAddressFromViewKey(const XmrPubKey& pubSpendKey, const XmrEllipticCurve& ec, const int account, const int index) const
{	
	if(isSpendKey)
	{
		throw std::string("Internal error, getSubAddressFromViewKey should only be used for private view key");
	}
	
	
	if(account == 0 && index == 0)
	{
		//std::cout << "This is (0, 0)!!!!!!!!!!!!!!!!!!!!!!!!!" << std::endl;
		return getStandardAddressView(pubSpendKey);
	}

	const BigInt<2048> a(key);
	
	const XmrECPoint B = pubSpendKey.point;

	
	const BigInt<32> accountInt(account);	
	const BigInt<32> indexInt(index);

	const std::string accountBuf = accountInt.getBinaryLong();
	const std::string indexBuf = indexInt.getBinaryLong();

	const std::string subAddrStr("SubAddr\0", 8);
			
	const std::string h2 = toBinaryXmrString();
	const std::string h3 = reverseStr(accountBuf);
	const std::string h4 =  reverseStr(indexBuf);
	
	const std::string binBuf = subAddrStr + h2 + h3 + h4;
			
	const std::string hashRaw = computeKeccak1600(binBuf);

	const std::string hash = reverseStr(hashRaw);
		
	const BigInt<256> m1(BigInt<256>::fromBytes(hash));
	const BigInt<256> m2 = XmrEllipticCurve::getEdwards25519Max(m1);			
	const BigInt<2048> m(m2);
		
		
	//D = B + m*G			
	const XmrECPoint m_G = ec.multiplyGenPoint(m);			
	const XmrECPoint D =  ec.addPoints(B, m_G);		

	const XmrPubKey publicSpendKey(D);

	//C = a*D
	const XmrECPoint C = ec.multiplyPoint(a, D);
	const XmrPubKey publicViewKey(C);

	const PubXmrAddress pubXmrAddress(publicSpendKey, publicViewKey, true);

	const std::string addrStr = pubXmrAddress.toString();
	
	return pubXmrAddress;
}


//Subaddress for private spend key
PubXmrAddress XmrPrivSpendKey::getSubAddress(const XmrEllipticCurve& ec, const int account, const int index) const
{
	if(!isSpendKey)
	{
		throw std::string("Internal error, getSubAddress should only be used for private spend key");
	}
	
	const XmrPrivViewKey privateViewKey = privateSpendToView();
	
	const XmrPubKey basePubSpendKey = getPubkey(ec);
	return privateViewKey.getSubAddressFromViewKey(basePubSpendKey, ec, account, index);
}


